﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NAudio.Wave;
using System.Runtime.InteropServices;
using System.IO;

namespace WpfAudioPlayer.Models.Wmf
{
	/// <summary>
	/// Windows Media Format Runtime を利用したオーディオ データの読み取り機能を提供します。
	/// このクラスの処理はインスタンスを生成したスレッドで完結する必要があります。
	/// </summary>
	class WmfReader : IDisposable
	{
		/// <summary>
		/// インスタンスを初期化します。
		/// </summary>
		/// <param name="fileName">ファイルへのパス。</param>
		public WmfReader( string fileName )
		{
			try
			{
				NativeMethods.WMCreateSyncReader( IntPtr.Zero, WMT_RIGHTS.WMT_RIGHT_PLAYBACK, out this._wmSyncReader );

				this._wmSyncReader.Open( fileName );
				this.Initialize();
			}
			catch( COMException exception )
			{
				if( ( ulong )exception.ErrorCode == NativeMethods.NS_E_FILE_OPEN_FAILED )
				{
					throw new OutOfMemoryException();
				}
				else
				{
					throw exception;
				}
			}
		}

		/// <summary>
		/// オーディオ出力を行う対象となる番号を取得します。
		/// </summary>
		/// <param name="reader">Windows Media オーディオの読み取りを行う為のオブジェクト。</param>
		/// <returns>オーディオ出力番号。</returns>
		private static uint GetAudioOutputNumber( IWMSyncReader reader )
		{
			uint outputNumber = WmfReader.InvalidOuput;
			uint count        = 0;

			reader.GetOutputCount( out count );
			for( uint index = 0; index < count; ++index )
			{
				IWMOutputMediaProps props = null;
				reader.GetOutputProps( index, out props );

				Guid mediaType;
				props.GetType( out mediaType );

				if( mediaType == WmfMediaTypes.WMMEDIATYPE_Audio )
				{
					outputNumber = index;
					break;
				}
			}

			return outputNumber;
		}

		/// <summary>
		/// オーディオ情報を取得する。
		/// </summary>
		/// <param name="reader">Windows Media オーディオの読み取りを行う為のオブジェクト。</param>
		/// <param name="outputNumber">オーディオ出力番号。</param>
		/// <param name="formatNumber">オーディオ フォーマット番号。</param>
		/// <returns></returns>
		private static WaveFormat GetOutputFormat( IWMSyncReader reader, uint outputNumber, uint formatNumber )
		{
			IWMOutputMediaProps props = null;
			reader.GetOutputFormat( outputNumber, formatNumber, out props );

			uint typeSize = 0;
			props.GetMediaType( IntPtr.Zero, ref typeSize );

			WaveFormat	waveFormat = null;
			IntPtr		buffer     = Marshal.AllocCoTaskMem( Math.Max( ( int )typeSize, Marshal.SizeOf( typeof( WM_MEDIA_TYPE ) ) + Marshal.SizeOf( typeof( WaveFormat ) ) ) );

			try
			{
				props.GetMediaType( buffer, ref typeSize );
				WM_MEDIA_TYPE mediaType =( WM_MEDIA_TYPE )Marshal.PtrToStructure( buffer, typeof( WM_MEDIA_TYPE ) );
				if( mediaType.MajorType  == WmfMediaTypes.WMMEDIATYPE_Audio &&
					mediaType.SubType    == WmfMediaTypes.WMMEDIASUBTYPE_PCM &&
					mediaType.FormatType == WmfMediaTypes.WMFORMAT_WaveFormatEx &&
					mediaType.FormatSize >= WmfReader.WaveFormatExSize )
				{
					waveFormat = WaveFormat.MarshalFromPtr( mediaType.Format );
				}
			}
			finally
			{
				Marshal.FreeCoTaskMem( buffer );
			}

			return waveFormat;
		}


		/// <summary>
		/// PCM 出力を行う対象となる番号を取得します。
		/// </summary>
		/// <param name="reader">オーディオ データ読み取りオブジェクト。</param>
		/// <param name="outputNumber">オーディオ出力番号のコレクション。</param>
		/// <returns>PCM 出力を行う対象となる番号。</returns>
		private static uint[] GetPCMOutputNumbers( IWMSyncReader reader, uint outputNumber )
		{
			List< uint > pcms = new List<uint>();

			uint formatCount = 0;
			reader.GetOutputFormatCount( outputNumber, out formatCount );

			int		bufferSize = Marshal.SizeOf( typeof( WM_MEDIA_TYPE ) ) + Marshal.SizeOf( typeof( WaveFormat ) );
			IntPtr	buffer = Marshal.AllocCoTaskMem( bufferSize );

			try
			{
				for( uint index = 0; index < formatCount; ++index )
				{
					IWMOutputMediaProps props = null;
					reader.GetOutputFormat( outputNumber, index, out props );

					uint typeSize = 0;
					props.GetMediaType( IntPtr.Zero, ref typeSize );

					if( ( int )typeSize > bufferSize )
					{
						bufferSize = ( int )typeSize;
						Marshal.FreeCoTaskMem( buffer );
						buffer = Marshal.AllocCoTaskMem( bufferSize );
					}

					props.GetMediaType( buffer, ref typeSize );

					WM_MEDIA_TYPE mediaType = ( WM_MEDIA_TYPE )Marshal.PtrToStructure( buffer, typeof( WM_MEDIA_TYPE ) );
					if( mediaType.MajorType  == WmfMediaTypes.WMMEDIATYPE_Audio &&
						mediaType.SubType    == WmfMediaTypes.WMMEDIASUBTYPE_PCM &&
						mediaType.FormatType == WmfMediaTypes.WMFORMAT_WaveFormatEx &&
						mediaType.FormatSize >= WmfReader.WaveFormatExSize )
					{
						pcms.Add( index );
					}
				}
			}
			finally
			{
				Marshal.FreeCoTaskMem( buffer );
			}

			return pcms.ToArray();
		}
		
		/// <summary>
		/// ストリームの初期化を行う。
		/// </summary>
		private void Initialize()
		{
			this._outputNumber = WmfReader.GetAudioOutputNumber( this._wmSyncReader );
			if( this._outputNumber == WmfReader.InvalidOuput )
			{
				throw new ArgumentException( "An audio stream was not found" );
			}

			uint[] formatIndexes = WmfReader.GetPCMOutputNumbers( this._wmSyncReader, this._outputNumber );
			if( formatIndexes.Length == 0 )
			{
				throw new ArgumentException( "An audio stream was not found" );
			}

			this._outputFormatNumber = formatIndexes[ 0 ];
			this._waveFormat = WmfReader.GetOutputFormat( this._wmSyncReader, this._outputNumber, this._outputFormatNumber );

			uint outputCount = 0;
			this._wmSyncReader.GetOutputCount( out outputCount );

			// ストリームの設定
			{
				WMT_STREAM_SELECTION[]	streamSelections = new WMT_STREAM_SELECTION[ outputCount ];
				ushort[]				streamNumbers   = new ushort[ outputCount ];

				for( uint index = 0; index < outputCount; index++ )
				{
					this._wmSyncReader.GetStreamNumberForOutput( index, out streamNumbers[ index ] );
					if( index == this._outputNumber )
					{
						streamSelections[ index ] = WMT_STREAM_SELECTION.WMT_ON;
						this._outputStream = streamNumbers[ index ];
						this._wmSyncReader.SetReadStreamSamples( this._outputStream, false );
					}
					else
					{
						streamSelections[ index ] = WMT_STREAM_SELECTION.WMT_OFF;
					}
				}

				this._wmSyncReader.SetStreamsSelected( ( ushort )outputCount, streamNumbers, streamSelections );
			}

			// 出力情報の設定
			{
				IWMOutputMediaProps props = null;
				this._wmSyncReader.GetOutputFormat( this._outputNumber, this._outputFormatNumber, out props );
				this._wmSyncReader.SetOutputProps( this._outputNumber, props );

				uint typeSize = 0;
				props.GetMediaType( IntPtr.Zero, ref typeSize );
				IntPtr buffer = Marshal.AllocCoTaskMem( ( int )typeSize );
				try
				{
					props.GetMediaType( buffer, ref typeSize );
					WM_MEDIA_TYPE mediaType = ( WM_MEDIA_TYPE )Marshal.PtrToStructure( buffer, typeof( WM_MEDIA_TYPE ) );
					this._sampleSize = mediaType.SampleSize;
				}
				finally
				{
					Marshal.FreeCoTaskMem( buffer );
					props = null;
				}
			}

			//
			{
				this._canSeek = false;
				this._length = -1;

				WmfHeaderInfo headerInfo = new WmfHeaderInfo( ( IWMHeaderInfo )this._wmSyncReader );
				try
				{
					WmfTagValue tagValue = headerInfo.GetTagValue( WmfAttributeNames.WMSeekable );
					if( tagValue != null ) { this._canSeek = tagValue.Boolean; }

					tagValue = headerInfo.GetTagValue( WmfAttributeNames.WMDuration );
					if( tagValue != null )
					{
						this._length = this.SampleTimeToBytePosition( tagValue.QWord );
						this._totalTime = new TimeSpan( ( long )tagValue.QWord );
					}
				}
				catch( COMException e )
				{
					if( e.ErrorCode != NativeMethods.ASF_E_NOTFOUND )
					{
						throw e;
					}
				}
			}
		}

		/// <summary>
		/// サンプル時間 ( 100 ナノ秒単位 ) をバイト単位のオーディオデータ上の位置情報へ変換します。
		/// </summary>
		/// <param name="sampleTime">サンプル時間。</param>
		/// <returns>バイト単位の位置情報。</returns>
		private long SampleTimeToBytePosition( ulong sampleTime )
		{
			var ret = sampleTime * ( ulong )this._waveFormat.AverageBytesPerSecond / 10000000;

			ret -= ( ret % ( ulong )_waveFormat.BlockAlign );
			return ( long )ret;
		}

		/// <summary>
		/// バイト単位のオーディオデータ上の位置情報をサンプル時間 ( 100 ナノ秒単位 ) へ変換します。
		/// </summary>
		/// <param name="pos">バイト単位の位置情報。</param>
		/// <returns>サンプル時間。</returns>
		private ulong BytePositionToSampleTime( long position )
		{
			position -= ( position % ( long )this._waveFormat.BlockAlign );
			return ( ulong )position * 10000000 / ( ulong )this._waveFormat.AverageBytesPerSecond;
		}

		/// <summary>
		/// 現在のストリーム内の位置を設定します。
		/// </summary>
		/// <param name="offset">origin パラメータからのバイト オフセット。</param>
		/// <param name="origin">新しい位置を取得するために使用する参照ポイントを示す SeekOrigin 型の値。</param>
		/// <returns>現在のストリーム内の新しい位置。</returns>
		public long Seek( long offset, SeekOrigin origin )
		{
			if( this.CanSeek )
			{
				switch( origin )
				{
				case SeekOrigin.Current:
					offset += this._position;
					break;

				case SeekOrigin.End:
					offset += this._length;
					break;
				}

				if( offset == this._position )
				{
					return this._position; // :-)
				}

				if( ( offset < 0 ) || ( offset > this._length ) )
				{
					throw new ArgumentException( "Offset out of range", "offset" );
				}

				if( ( offset % this.SeekAlign ) > 0 )
				{
					throw new ArgumentException( string.Format( "Offset must be aligned by a value of SeekAlign ({0})", this.SeekAlign ), "offset" );
				}

				ulong sampleTime = this.BytePositionToSampleTime( offset );
				this._wmSyncReader.SetRange( sampleTime, 0 );

				this._position = offset;
				this._wmBufferReader = null;

				return offset;
			}
			else
			{
				throw new NotSupportedException();
			}
		}

		/// <summary>
		/// オーディオ データを読み取ります。
		/// </summary>
		/// <param name="buffer">読み取ったデータを格納するバッファ。</param>
		/// <param name="offset">読み取り位置。</param>
		/// <param name="count">読み取るサイズ。</param>
		/// <returns>読み取りに成功したサイズ。</returns>
		public int Read( byte[] buffer, int offset, int count )
		{
			if( this._wmSyncReader == null ) { throw new ObjectDisposedException( this.ToString() ); }

			int read = 0;
			if( ( this._length > 0 ) && ( ( this._length - this._position ) < count ) )
			{
				count = ( int )( this._length - this._position );
			}

			if( this._wmBufferReader != null )
			{
				while( ( this._wmBufferReader.Position < this._wmBufferReader.Length ) && ( read < count ) )
				{
					read += this._wmBufferReader.Read( buffer, offset, count );
				}
			}

			while( read < count )
			{
				INSSBuffer	sample    = null;
				ulong		sampleTime = 0;
				ulong		duration   = 0;
				uint		flags      = 0;

				try
				{
					this._wmSyncReader.GetNextSample( this._outputStream, out sample, out sampleTime, out duration, out flags, out this._outputNumber, out this._outputStream );
				}
				catch( COMException e )
				{
					if( e.ErrorCode == NativeMethods.NS_E_NO_MORE_SAMPLES )
					{
						if( this._waveFormat.BitsPerSample == 8 )
						{
							while( read < count )
							{
								buffer[ offset + read ] = 0x80;
								++read;
							}
						}
						else
						{
							Array.Clear( buffer, offset + read, count - read );
							read = count;
						}
						break;
					}
					else
					{
						throw ( e );
					}
				}

				this._wmBufferReader = new WmfNSSBuffer( sample );
				read += this._wmBufferReader.Read( buffer, offset + read, count - read );
			}

			if( ( this._wmBufferReader != null ) && ( this._wmBufferReader.Position >= this._wmBufferReader.Length ) )
			{
				this._wmBufferReader = null;
			}

			this._position += read;
			return read;
		}
		
		/// <summary>
		/// シーク操作をサポートしている事を示す値を取得します。
		/// </summary>
		public bool CanSeek
		{
			get
			{
				return this._canSeek;
			}
		}

		/// <summary>
		/// 現在の音声データ中の位置を時間単位で取得または設定する。
		/// </summary>
		public TimeSpan CurrentTime
		{
			get
			{
				return TimeSpan.FromSeconds( ( double )this.Position / this.WaveFormat.AverageBytesPerSecond );
			}
			set
			{
				this.Position = ( long )( value.TotalSeconds * this.WaveFormat.AverageBytesPerSecond );
			}
		}

		/// <summary>
		/// オーディオ データの長さを取得します。
		/// </summary>
		public long Length
		{
			get
			{
				return this._length;
			}
		}

		/// <summary>
		/// オーディオ データ内の位置を取得または設定します。
		/// </summary>
		public long Position
		{
			get
			{
				return this._position;
			}
			set
			{
				this.Seek( value, SeekOrigin.Begin );
			}
		}
		
		/// <summary>
		/// シーク操作に必要なデータ長を取得します。
		/// </summary>
		private long SeekAlign
		{
			get
			{
				return Math.Max( this.SampleTimeToBytePosition( 1 ), ( long )this._waveFormat.BlockAlign );
			}
		}

		/// <summary>
		/// 総演奏時間を取得します。
		/// </summary>
		public TimeSpan TotalTime
		{
			get
			{
				return this._totalTime;
			}
		}	
		
		/// <summary>
		/// オーディオ情報を取得します。
		/// </summary>
		public WaveFormat WaveFormat
		{
			get
			{
				return this._waveFormat;
			}
		}

		#region フィールド

		/// <summary>
		/// WAVE_FORMAT_EX 構造体のサイズ。
		/// </summary>
		private const uint WaveFormatExSize = 18;

		/// <summary>
		/// 不正なオーディオ出力番号。
		/// </summary>
		private const uint InvalidOuput = 0xFFFFFFFF;
	
		/// <summary>
		/// シーク操作をサポートしている事を示す値。
		/// </summary>
		private bool _canSeek;

		/// <summary>
		/// ストリームの長さ ( バイト単位 )。
		/// </summary>
		private long _length;
		
		/// <summary>
		/// オーディオ出力番号。
		/// </summary>
		private uint _outputNumber;

		/// <summary>
		/// オーディオ出力のフォーマット番号。
		/// </summary>
		private uint _outputFormatNumber;

		/// <summary>
		/// オーディオ出力ストリーム。
		/// </summary>
		private ushort _outputStream;

		/// <summary>
		/// 現在のストリーム内の位置。
		/// </summary>
		private long _position = 0;

		/// <summary>
		/// サンプリング サイズ。
		/// </summary>
		private uint _sampleSize;

		/// <summary>
		/// 総演奏時間。
		/// </summary>
		private TimeSpan _totalTime;

		/// <summary>
		/// オーディオ情報。
		/// </summary>
		private WaveFormat _waveFormat;
		
		/// <summary>
		/// オーディオ データのバッファ。
		/// </summary>
		private WmfNSSBuffer _wmBufferReader = null;

		/// <summary>
		/// Windows Media オーディオの読み取りを行う為のオブジェクト。
		/// </summary>
		private IWMSyncReader _wmSyncReader;

		#endregion

		#region IDisposable メンバ

		/// <summary>
		/// リソースの解放を行います。
		/// </summary>
		public void Dispose()
		{
			if( this._wmSyncReader != null )
			{
				this._wmSyncReader.Close();
				this._wmSyncReader = null;
			}
		}

		#endregion
	}
}
